import collections
import logging
import re
import socketserver
import threading
import time

from tcpnetlock import utils

"""
This implement a very simple network lock server based on just TCP.

The server listen for connections, reads the socket for a '\n' terminated line.
The received string is used as lock name.

If the lock is granted, it is valid until the client release it, or until the TCP connection is closed.

If the lock can't be granted, a response is sent to the client and the TCP connection is closed.


Detail of the current implementation:

    1. SERVER accept TCP connection
    2. CLIENT send the lock name (utf-8 encoded), like: 'lock-name\n'
    3. SERVER tries to acquire the lock
        a. if lock was granted, returns 'ok\n' to the client and the TCP connections is KEPT OPEN
            - the LOCK remains granted until:
               + the client release it, sending 'release\n' to the server (and the TCP connection is CLOSED)
               + the client closes the TCP connection
               + for some reason the TCP connection is lost
        b. if lock was NOT granted, returns 'failed\n' to the client and the TCP connection is immediately CLOSED
"""

logger = logging.getLogger(__name__)


RESPONSE_OK = 'ok'
RESPONSE_ERR = 'err'
RESPONSE_LOCK_FAILED = 'failed'
RESPONSE_RELEASED = 'released'
RESPONSE_SHUTTING_DOWN = 'shutting-down'
RESPONSE_PONG = 'pong'
RESPONSE_STILL_ALIVE = 'alive'

ACTION_RELEASE = 'release'
ACTION_SERVER_SHUTDOWN = '.server-shutdown'
ACTION_PING = '.ping'
ACTION_KEEPALIVE = '.keepalive'

VALID_LOCK_NAME_RE = re.compile(r'^[a-zA-Z0-9_-]+$')
VALID_CLIENT_ID_RE = re.compile(r'^[a-zA-Z0-9_-]+$')


class Holder:
    def __init__(self):
        self._lock = self._lock = threading.Lock()
        self._lock_name = None
        self._timestamp = None
        self._client_id = None

    def try_acquire(self, lock_name, client_id):
        granted = self._lock.acquire(blocking=False)
        if granted:
            self._timestamp = time.time()
            self._client_id = client_id
            self._lock_name = lock_name
        return granted

    def release(self):
        self._lock.release()

    def __str__(self):
        return "Lock '{name}', client '{client}', age: {age}".format(
            name=self._lock_name, client=self._client_id, age=int(time.time() - self._timestamp))


class LockServer(socketserver.ThreadingTCPServer):
    allow_reuse_address = True

    def __init__(self, host='localhost', port=9999):
        super().__init__((host, port), TCPHandler)


class TCPHandler(socketserver.BaseRequestHandler):
    GLOBAL_LOCK = threading.Lock()
    LOCKS = collections.defaultdict(Holder)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._lock_name = None

    def _get_line(self) -> str:
        """
        Reads (blocking) until we got a lock name
        :return: lock name
        """
        binary_data = bytearray()
        while True:
            lock_name = utils.try_get_line(self.request, binary_data)
            if lock_name:
                return lock_name

    def _handle_lock(self, holder: Holder):
        try:
            try:
                self._real_handle_lock(holder)
            except utils.ClientDisconnected:
                logger.info("[%s] Client disconnected", self._lock_name)
        finally:
            holder.release()

    def _send(self, message):
        self.request.send((message + '\n').encode())

    def _real_handle_lock(self, holder: Holder):
        """
        While control is in this method, the lock is held
        """
        self._send(RESPONSE_OK)
        binary_data = bytearray()
        while True:
            action = utils.try_get_line(self.request, binary_data, timeout=1.0)
            if action:
                logger.info("Action: '%s' for lock %s", action, holder)
                binary_data = bytearray()
                if action == ACTION_RELEASE:
                    logger.debug("Releasing lock: %s", holder)
                    self._send(RESPONSE_RELEASED)
                    self.request.close()
                    # FIXME: at this point we send OK to the client, but internally the lock is STILL HELD
                    return
                if action == ACTION_KEEPALIVE:
                    logger.debug("Received keepalive from client. Lock: %s", holder)
                    self._send(RESPONSE_STILL_ALIVE)
                else:
                    logger.debug("Unknown action '%s' for lock: %s", action, holder)

    def _handle_server_shutdown(self):
        # FIXME: assert connections came from localhost
        self._send(RESPONSE_SHUTTING_DOWN)  # FIXME: what if client had closed the socket?
        self.request.close()
        self.server.shutdown()

    def _handle_ping(self):
        self._send(RESPONSE_PONG)  # FIXME: what if client had closed the socket?
        self.request.close()

    def _handle_invalid_lock_hame(self):
        self._send(RESPONSE_ERR + ':invalid lock name')  # FIXME: what if client had closed the socket?
        self.request.close()

    def handle(self):
        try:
            line = self._get_line()
        except utils.ClientDisconnected:
            logger.info("Client disconnected")
            self.request.close()
            return

        logger.debug("Received line: '%s'", line)

        if line == ACTION_SERVER_SHUTDOWN:
            self._handle_server_shutdown()
            return

        if line == ACTION_PING:
            self._handle_ping()
            return

        # Not a special string? Then it's a lock name (plus optional client-id)
        tokens = line.split(':')
        try:
            (self._lock_name, client_id) = tokens
        except ValueError:
            (self._lock_name, client_id) = tokens[0], None

        if not VALID_LOCK_NAME_RE.match(self._lock_name):
            self._handle_invalid_lock_hame()
            return

        # Get the holder
        self.GLOBAL_LOCK.acquire()
        try:
            lock_holder = self.LOCKS[self._lock_name]
        finally:
            self.GLOBAL_LOCK.release()

        # Lock it
        granted = lock_holder.try_acquire(self._lock_name, client_id)
        if granted:
            logger.debug("Got lock :)")
            self._handle_lock(lock_holder)
        else:
            logger.debug("Couldn't get lock :(")
            self._send(RESPONSE_LOCK_FAILED)  # FIXME: what if client had closed the socket?

        self.request.close()


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("--listen", default='localhost')
    parser.add_argument("--port", default=9999, type=int)
    parser.add_argument("--debug", default=False, action='store_true')
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    with LockServer(args.listen, args.port) as server:
        logging.info("Started listening on %s:%s", args.listen, args.port)
        server.serve_forever()


if __name__ == '__main__':
    main()



import logging
import socket
import sys
import time

from tcpnetlock import server
from tcpnetlock import utils

logger = logging.getLogger(__name__)


class LockClient:
    """
    This is a temporal implementation of the client, with just the basic stuff to be able to develop
    unittests for the server code. Some of this code may be used, in the future, for the implementation
    of the real client.
    """

    def __init__(self, host='localhost', port=9999, client_id=None, print_marks=False):
        """
        Creates a client to connect to the server.

        :param host: hostname of the server
        :param port: port to connect
        """
        self._host = host
        self._port = port
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._acquired = None
        self._print_marks = bool(print_marks)
        if client_id:
            assert self.valid_client_id(client_id), "Invalid client_id: {client_id}".format(client_id=client_id)
        self._client_id = client_id

    @staticmethod
    def valid_lock_name(lock_name):
        """Returns True if the provided lock name is valid"""
        return bool(server.VALID_LOCK_NAME_RE.match(lock_name))

    @staticmethod
    def valid_client_id(client_id):
        """Returns True if the provided client_id is valid"""
        return bool(server.VALID_CLIENT_ID_RE.match(client_id))

    @staticmethod
    def _assert_response(response: str, valid_response_codes):
        response_code = response.split(":")[0]
        assert response_code in valid_response_codes,\
            "Invalid response: '{response}'. Valid responses: {valid_response_codes}".format(
                response=response, valid_response_codes=valid_response_codes
            )

    def _send(self, message: str):
        assert message
        self._socket.send((message + '\n').encode())

    def _read_response(self, valid_responses):
        # FIXME: we should time-out here after some time
        response = utils.get_line(self._socket)
        self._assert_response(response, valid_responses)
        return response

    def connect(self):
        """Connects to the server"""
        logger.info("Connecting to '%s:%s'...", self._host, self._port)
        try:
            self._socket.connect((self._host, self._port))
            self._mark('CONNECTION,OK')
        except ConnectionRefusedError:
            self._mark('CONNECTION,REFUSED')
            raise
        except:  # noqa: E722
            self._mark('CONNECTION,ERROR')
            raise

    def _mark(self, message, *args, **kwargs):
        """Print easy to parse messages"""
        if self._print_marks:
            logging.debug("MARK," + message + ",END", *args, **kwargs)

    def lock(self, name: str) -> bool:
        """
        Tries to acquire a lock
        :param name: lock name
        :return: boolean indicating if lock as acquired or not
        """
        assert self.valid_lock_name(name)
        logger.info("Trying to acquire lock '%s'...", name)
        if self._client_id:
            message = "{name}:{client_id}".format(name=name, client_id=self._client_id)
        else:
            message = name
        self._send(message)
        response = self._read_response([server.RESPONSE_OK, server.RESPONSE_LOCK_FAILED, server.RESPONSE_ERR])
        self._acquired = (response == server.RESPONSE_OK)
        logging.info("Lock %s acquired: %s", name, self._acquired)
        self._mark("LOCK_RESULT,LOCK:%s,ACQUIRED:%s", name, self._acquired)

        return self._acquired

    def server_shutdown(self):
        """Send order to shutdown the server"""
        logger.info("Sending SHUTDOWN...")
        self._send(server.ACTION_SERVER_SHUTDOWN)
        return self._read_response([server.RESPONSE_SHUTTING_DOWN])

    def ping(self):
        """Send ping to the server"""
        logger.info("Sending PING...")
        self._send(server.ACTION_PING)
        return self._read_response([server.RESPONSE_PONG])

    def keepalive(self):
        """Send a keepalive to the server"""
        logger.info("Sending KEEPALIVE...")
        self._send(server.ACTION_KEEPALIVE)
        response = self._read_response([server.RESPONSE_STILL_ALIVE])
        self._mark('KEEPALIVE,OK')
        return response

    def release(self):
        """Release the held lock"""
        logger.info("Trying to RELEASE...")
        self._send(server.ACTION_RELEASE)
        return self._read_response([server.RESPONSE_RELEASED])

    def close(self):
        """Close the socket. As a result of the disconnection, the lock will be released at the server."""
        self._socket.close()

    @property
    def acquired(self) -> bool:
        """
        Returns boolean indicating if lock was acquired or not
        :return: True if lock was acquired, False otherwise
        """
        assert self._acquired in (True, False)  # Fail if lock() wasn't called
        return self._acquired


def main():
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument("lock_name")
    parser.add_argument("--host", default='localhost')
    parser.add_argument("--port", default=9999, type=int)
    parser.add_argument("--client-id", default=None)
    parser.add_argument("--keep-alive", default=False, action='store_true')
    parser.add_argument("--keep-alive-secs", default=15, type=int)
    parser.add_argument("--debug", default=False, action='store_true')
    parser.add_argument("--print-marks", default=False, action='store_true')
    args = parser.parse_args()

    if args.debug or args.print_marks:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    client = LockClient(args.host, args.port, client_id=args.client_id, print_marks=args.print_marks)
    client.connect()
    granted = client.lock(args.lock_name)
    if not granted:
        logger.debug("Lock not granted. Exiting...")
        sys.exit(9)

    if args.keep_alive:
        while True:
            logger.debug("Sleeping for %s... (after that, will send a keep-alive)", args.keep_alive_secs)
            time.sleep(args.keep_alive_secs)
            client.keepalive()
    else:
        while True:
            logger.debug("Sleeping for an hour...")
            time.sleep(60 * 60)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `tcpnetlock` package."""
import threading
import time
import uuid
from unittest import mock

import pytest

from tcpnetlock.client import LockClient
from tcpnetlock.server import LockServer


def _wait_for_server():
    """
    Busy-waits until the test server is responding
    """
    client = LockClient()
    for _ in range(10):
        try:
            client.connect()
            break
        except ConnectionRefusedError:
            time.sleep(0.1)
    client.ping()
    client.close()


@pytest.fixture(scope="module")
def lock_server():
    """
    Fixture, returns the server process running a LockServer ready to use
    """
    def start_server():
        server = LockServer("localhost", 9999)
        server.serve_forever()

    server_thread = threading.Thread(target=start_server, daemon=True)
    server_thread.start()
    _wait_for_server()  # wait until server is ready

    yield server_thread

    client = LockClient()
    client.connect()
    client.server_shutdown()
    client.close()

    server_thread.join(5)
    assert not server_thread.is_alive(), "TEST Server didn't shut down cleanly"


def test_server_is_alive(lock_server):
    """Test the test server is actually running"""
    assert lock_server.is_alive()
    client = LockClient()
    client.connect()
    client.ping()


def test_connect_and_get_lock_works(lock_server):
    """Test that a lock can be acquired"""
    client = LockClient()
    client.connect()
    acquired = client.lock(uuid.uuid4().hex)
    assert acquired
    client.close()


def test_connect_and_get_lock_with_client_id_works(lock_server):
    """Test that a lock can be acquired"""
    name = uuid.uuid4().hex
    client_id = uuid.uuid4().hex

    client = LockClient(client_id=client_id)
    client.connect()
    acquired = client.lock(name)
    assert acquired
    client.close()


def test_connect_and_get_lock_works(lock_server):
    """Test that a lock can be acquired"""
    client = LockClient()
    client.connect()
    acquired = client.lock(uuid.uuid4().hex)
    for _ in range(5):
        client.keepalive()
        time.sleep(0.5)
    client.close()


def test_get_two_different_lock(lock_server):
    """Test that two different locks can be acquired"""
    client_1 = LockClient()
    client_1.connect()
    acquired = client_1.lock(uuid.uuid4().hex)
    assert acquired
    client_1.close()

    client_2 = LockClient()
    client_2.connect()
    acquired = client_2.lock(uuid.uuid4().hex)
    assert acquired
    client_2.close()


def test_lock_twice_fails(lock_server):
    """Test that a locks can NOT be acquired twice"""
    name = uuid.uuid4().hex

    client_1 = LockClient()
    client_1.connect()
    acquired = client_1.lock(name)
    assert acquired

    client_2 = LockClient()
    client_2.connect()
    acquired = client_2.lock(name)
    assert not acquired

    client_1.close()
    client_2.close()


def test_released_lock_and_be_re_acquired(lock_server):
    """Test that a locks can be acquired again after it's released"""
    name = uuid.uuid4().hex

    # acquire lock
    client_1 = LockClient()
    client_1.connect()
    acquired = client_1.lock(name)
    assert acquired
    # release lock
    client_1.release()
    client_1.close()

    # re-acquire same lock
    client_2 = LockClient()
    client_2.connect()
    acquired = client_2.lock(name)
    assert acquired
    client_2.close()


def test_lock_released_when_client_closes_connection(lock_server):
    """Test that a locks can be acquired again after it's released"""
    name = uuid.uuid4().hex

    # acquire lock
    client_1 = LockClient()
    client_1.connect()
    acquired = client_1.lock(name)
    assert acquired
    # don't release socket, just close the connection
    client_1.close()

    # re-acquire same lock
    acquired = False
    for _ in range(20):
        client_2 = LockClient()
        client_2.connect()
        acquired = client_2.lock(name)
        if not acquired:
            time.sleep(0.1)
    assert acquired


def test_server_rejects_invalid_lock_name(lock_server):
    """Test that a locks can be acquired again after it's released"""
    invalid_names = (
        '.starts-with-space',
        'contains space',
        'contains%invalid%chars',
    )

    for invalid in invalid_names:
        client = LockClient()
        client.valid_lock_name = mock.MagicMock(return_value=True)
        client.connect()
        acquired = client.lock(invalid)
        assert not acquired, "Lock granted for invalid lock name: '{invalid}'".format(invalid=invalid)
        client.close()


def test_server_accept_valid_lock_name(lock_server):
    valid_names = (
        'valid-name',
        'valid-name-8',
        'valid_name',
        ' space_is_ignored',
        'space_is_ignored ',
        ' space_is_ignored ',
    )

    for valid in valid_names:
        client = LockClient()
        client.valid_lock_name = mock.MagicMock(return_value=True)
        client.connect()
        acquired = client.lock(valid)
        assert acquired, "Lock NOT granted for valid lock name: '{valid}'".format(valid=valid)
        client.close()



import logging
import socket


logger = logging.getLogger(__name__)


def get_line(sock, timeout=30):
    binary_data = bytearray()
    # FIXME: this way of implementing timeout is easy and simple, but not very accurate
    # since_try_get_line() will return earlier than 1 sec.
    for _ in range(timeout):
        line = try_get_line(sock, binary_data, timeout=1)
        if line:
            return line
    return None


class ClientDisconnected(Exception):
    pass


def try_get_line(sock: socket.socket, binary_data: bytearray, timeout=None):
    """
    * timeout=None -> disable timeout and red (will BLOCK until we get data)
    * timeout=x -> set timeout of x and red
    """
    # We need 'recv_size = 1' because we don't handle (yet) the case were client sends 'lock1\nrelease\n'
    # FIXME: re-implement parser, in way that handles correctly a message like 'lock\nrelease\n' with recv(16)
    recv_size = 1

    if timeout:
        logger.debug('Setting socket timeout to %s', timeout)
        sock.settimeout(timeout)
    else:
        logger.debug('Disabling socket timeout (will block)')
        sock.settimeout(None)

    logger.debug('Reading data...')
    try:
        rr = sock.recv(recv_size)
    except socket.timeout:
        rr = None
    else:
        if not rr:
            logger.debug("Client disconnected")
            raise ClientDisconnected()

    logger.debug("Data received: %s", rr)
    if rr:
        binary_data.extend(rr)

    logger.debug("binary_data: '%s'", binary_data)
    try:
        # FIXME: assuming decode() on partial data works and is safe on utf-8. IS THIS TRUE? Use ASCII otherwise
        potential_line = binary_data.decode()
    except UnicodeDecodeError:
        logger.debug("decoded failed for binary_data '%s'", binary_data)
        return None

    if potential_line.endswith('\n'):
        assert len(potential_line.splitlines()) == 1
        return potential_line.strip()

    return None
